﻿Imports UPLib
Imports System.Drawing
Imports System.Drawing.Imaging

Module Module1
    Dim JPG As Boolean
    Dim JPGQuality As Long = 95
    Dim OGG As Boolean

    Private Function InMyDir(path As String) As String
        Return IO.Path.Combine(My.Application.Info.DirectoryPath, path)
    End Function

    Private Sub WriteRed(msg As String)
        Console.ForegroundColor = ConsoleColor.Red
        Console.WriteLine(msg)
        Console.ResetColor()
    End Sub

    Private Function GetEncoder(format As ImageFormat) As ImageCodecInfo
        Dim codecs = ImageCodecInfo.GetImageDecoders()
        For Each codec In codecs
            If codec.FormatID = format.Guid Then
                Return codec
            End If
        Next
        Return Nothing
    End Function

    Sub SaveJPG(bmp As Bitmap, file As String)
        Dim jpgEncoder = GetEncoder(ImageFormat.Jpeg)
        Dim myEncoder = System.Drawing.Imaging.Encoder.Quality
        Dim myEncoderParameters = New EncoderParameters(1)
        Dim myEncoderParameter = New EncoderParameter(myEncoder, JPGQuality)
        myEncoderParameters.Param(0) = myEncoderParameter
        bmp.Save(file, jpgEncoder, myEncoderParameters)
    End Sub

    Private Function GetTempFile(extension As String) As String
        Dim file = "temp" & extension
        Dim i = 1
start:
        If IO.File.Exists(file) Then
            i += 1
            file = "temp" & i & extension
            GoTo start
        End If

        Return file
    End Function

    Sub Main()
        Dim filesToExtract As New List(Of String)

        For Each arg In My.Application.CommandLineArgs
            If IO.File.Exists(arg) Then
                filesToExtract.Add(arg)
            Else
                If arg.ToLower = "-jpg" Then JPG = True
                If arg.ToLower = "-ogg" Then OGG = True
                If arg.ToLower.StartsWith("-jpg-quality=") Then JPGQuality = arg.Substring(13)
            End If
        Next

        If filesToExtract.Count = 0 Then
            Console.WriteLine("UnExtract" & vbCrLf & "by Yrex, 2021" & vbCrLf & vbCrLf & "Usage:" & vbCrLf & "Drag and drop one or more Unreal packages on the icon. The following assets will be extracted: BSP and actors, Texture, FireTexture, WetTexture, Font, Sound." & vbCrLf & vbCrLf & "Advanced usage:" & vbCrLf & "unextract [options] input [input2...]" & vbCrLf & "-jpg: Save opaque textures as JPG. May break masking." & vbCrLf & "-jpg-quality=num: Change JPG quality (0-100, default is 95)" & vbCrLf & "-ogg: Save all sounds as OGG (instead of WAV). There shouldn't be a noticeable  quality loss.")
            Console.WriteLine("Press any key to close...")
            Console.ReadKey(True)
            Return
        End If

        For Each file In filesToExtract
            My.Computer.FileSystem.CurrentDirectory = IO.Path.GetDirectoryName(file)
            ExtractPackage(file)
        Next

        Console.ForegroundColor = ConsoleColor.Green
        Console.WriteLine("Finished.")
        Console.ResetColor()
        Console.WriteLine("Press any key to close...")
        Console.ReadKey(True)
    End Sub

    Sub ExtractPackage(file As String)
        Console.WriteLine("Extracting " & file & "...")

        Dim outDir = "."

        Dim pkg As New Package(file)

        Dim pkgName = IO.Path.GetFileNameWithoutExtension(file)

        For i = 0 To pkg.ExportTable.Count - 1
            Dim obj = pkg.ExportTable(i)

            If Not NameEq(obj.ObjectNameStr, "None") Then
                Select Case obj.ClassNameStr
                    Case "Texture", "Sound", "FireTexture", "WetTexture", "Font"
                        Console.WriteLine("Exporting " & pkg.GetFullObjectName(i + 1, True) & "...")
                        If Not IO.Directory.Exists(pkgName) Then
                            IO.Directory.CreateDirectory(pkgName)
                        End If
                End Select

                Select Case obj.ClassNameStr
                    Case "Texture"
                        Dim tex = Textures.GetTexture(obj)
                        Dim bmp = tex.ToBitmap(, True)
                        Dim groupPath = ""
                        Dim group = obj.Package
                        If group <> 0 Then groupPath = pkg.GetShortObjectName(group) & "\"
                        IO.Directory.CreateDirectory(pkgName & "\Textures\" & groupPath)
                        Dim path = IO.Path.Combine(outDir, pkgName & "\Textures\" & groupPath & obj.ObjectNameStr)
                        If Not JPG Or tex.bMasked Or tex.Props.GetBool("bAlphaBlend") Then
                            bmp.Save(path & ".png", System.Drawing.Imaging.ImageFormat.Png)
                        Else
                            SaveJPG(bmp, path & ".jpg")
                        End If
                        SaveTexProps(obj, path & ".ini")

                    Case "FireTexture"
                        Dim groupPath = ""
                        Dim group = obj.Package
                        If group <> 0 Then groupPath = pkg.GetShortObjectName(group) & "\"
                        IO.Directory.CreateDirectory(pkgName & "\Textures\" & groupPath)
                        Dim path = IO.Path.Combine(outDir, pkgName & "\Textures\" & groupPath & obj.ObjectNameStr)
                        SaveFireTexture(obj, path & ".firet")
                        SaveTexProps(obj, path & ".ini")

                    Case "WetTexture"
                        Dim groupPath = ""
                        Dim group = obj.Package
                        If group <> 0 Then groupPath = pkg.GetShortObjectName(group) & "\"
                        IO.Directory.CreateDirectory(pkgName & "\Textures\" & groupPath)
                        Dim path = IO.Path.Combine(outDir, pkgName & "\Textures\" & groupPath & obj.ObjectNameStr)
                        SaveWetTexture(obj, path & ".wett")
                        SaveTexProps(obj, path & ".ini")

                    Case "Font"
                        Dim groupPath = ""
                        Dim group = obj.Package
                        If group <> 0 Then groupPath = pkg.GetShortObjectName(group) & "\"
                        IO.Directory.CreateDirectory(pkgName & "\Textures\" & groupPath)
                        Dim path = IO.Path.Combine(outDir, pkgName & "\Textures\" & groupPath & obj.ObjectNameStr)
                        SaveFont(obj, path & ".font")

                    Case "Sound"
                        Dim groupPath = ""
                        Dim group = obj.Package
                        If group <> 0 Then groupPath = pkg.GetShortObjectName(group) & "\"
                        IO.Directory.CreateDirectory(pkgName & "\Sounds\" & groupPath)
                        Dim path = IO.Path.Combine(outDir, pkgName & "\Sounds\" & groupPath & obj.ObjectNameStr)
                        If OGG Then
                            Dim wavFile = GetTempFile(".wav")
                            Dim oggFile = GetTempFile(".ogg")

                            IO.File.WriteAllBytes(wavFile, GetSound(obj))
                            Shell("""" & InMyDir("oggenc.exe") & """ -o " & oggFile & " " & wavFile, , True)
                            IO.File.Delete(wavFile)
                            IO.File.Move(oggFile, path & ".ogg")
                        Else
                            IO.File.WriteAllBytes(path & ".wav", GetSound(obj))
                        End If
                End Select
            End If
        Next

        If IsALevel(pkg) Then
            Console.WriteLine("Exporting the level...")
            SaveLevel(pkg, pkgName & ".t3d", pkgName & ".bsp")
        End If
    End Sub

    Sub SaveTexProps(tex As ExportTableItem, file As String)
        Dim r = tex.Data.R
        r.BaseStream.Position = 0
        Dim props = tex.GetProperties()

        Dim w As IO.StreamWriter = Nothing

        For i = 0 To props.Props.Count - 1
            Dim p = props.Props(i)
            Select Case p.NameStr.ToLower()
                Case "palette", "usize", "vsize", "ubits", "vbits", "uclamp", "vclamp", "mipzero", "maxcolor", "internaltime"
                Case Else
                    If w Is Nothing Then w = New IO.StreamWriter(file)
                    w.WriteLine(p.NameStr & "=" & props.GetValueStr(i))
            End Select
        Next

        If w IsNot Nothing Then
            w.Close()
        End If
    End Sub

    Sub SaveFireTexture(ft As ExportTableItem, file As String)
        Dim r = ft.Data.R
        r.BaseStream.Position = 0
        Dim props = ft.GetProperties()

        Using w As New IO.BinaryWriter(New IO.FileStream(file, IO.FileMode.Create, IO.FileAccess.Write))
            w.Write(CInt(1)) 'type (FireTexture)
            w.Write(props.GetInt("USize"))
            w.Write(props.GetInt("VSize"))

            Dim palIndex = props.GetInt("Palette")
            Dim pal = GetPalette(ft.Pkg.ExportTable(palIndex - 1))
            For i = 0 To 256 - 1
                If i < pal.Length Then
                    w.Write(CInt(pal(i).ToArgb))
                Else
                    w.Write(CInt(&HFF000000))
                End If
            Next

            w.Write(props.GetBool("bRising"))
            w.Write(CByte(props.GetInt("RenderHeat")))
            w.Write(props.GetInt("SparksLimit"))

            r.ReadByte()
            If ft.Pkg.PackageVersion >= 63 Then
                r.ReadInt32()
            End If
            r.ReadCompact()
            r.ReadInt32()
            r.ReadInt32()
            r.ReadByte()
            r.ReadByte()

            Dim numSparks = r.ReadCompact()
            w.Write(CInt(numSparks))

            w.Write(r.ReadBytes(8 * numSparks))
        End Using
    End Sub

    Sub SaveWetTexture(wt As ExportTableItem, file As String)
        Dim r = wt.Data.R
        r.BaseStream.Position = 0
        Dim props = wt.GetProperties()

        Using w As New UnBinaryWriter(New IO.FileStream(file, IO.FileMode.Create, IO.FileAccess.Write))
            w.WriteInt32(2) 'type (WetTexture)
            w.WriteInt32(props.GetInt("USize"))
            w.WriteInt32(props.GetInt("VSize"))

            Dim srcTexName = wt.Pkg.GetFullObjectName(props.GetInt("SourceTexture"), False)
            w.WriteZString(srcTexName)

            w.WriteByte(props.GetInt("WaveAmp"))

            Dim numDrops = 0
            For Each prop In props.Props
                If StrEq(prop.NameStr, "Drops") Then
                    numDrops += 1
                End If
            Next

            w.WriteInt32(numDrops)
            For Each prop In props.Props
                If StrEq(prop.NameStr, "Drops") Then
                    If prop.RawValue.Length <> 8 Then
                        Throw New Exception("sizeof(Drops[0])<>8")
                    End If
                    w.Write(prop.RawValue)
                End If
            Next
        End Using
    End Sub

    Sub SaveFont(font As ExportTableItem, file As String)
        ' This isn't perfect, but should work most of the time.
        If SaveFontInternal(font, file, 1) = False Then
            If SaveFontInternal(font, file, 0) = False Then
                Console.ForegroundColor = ConsoleColor.Red
                Console.WriteLine("Can't export font")
                Console.ResetColor()
            End If
        End If
    End Sub

    Private Function SaveFontInternal(font As ExportTableItem, file As String, hack As Integer) As Boolean
        Dim r = font.Data.R
        r.BaseStream.Position = 0
        Try
            Dim props = font.GetProperties()

            If hack Then
                r.ReadByte() 'unknown
            End If

            Dim numPages = r.ReadCompact
            If numPages = 0 Or numPages > 255 Then Return False

            Using w As New UnBinaryWriter(New IO.FileStream(file, IO.FileMode.Create, IO.FileAccess.Write))
                w.WriteInt32(16) 'type (Font)

                w.WriteInt32(numPages)
                For p = 0 To numPages - 1

                    Dim texIndex = r.ReadCompact
                    If texIndex < 0 Then Return False
                    If texIndex > 0 Then
                        If Not NameEq(font.Pkg.ExportTable(texIndex - 1).ClassNameStr, "Texture") Then
                            Return False
                        End If
                    End If
                    Dim texName = font.Pkg.GetFullObjectName(texIndex, False)
                    w.WriteZString(texName)

                    Dim numCharacters = r.ReadCompact
                    If numCharacters > 255 Then Return False
                    w.WriteInt32(numCharacters)
                    w.Write(r.ReadBytes(numCharacters * 4 * 4))
                Next
            End Using
        Catch
            Return False
        End Try
        Return True
    End Function

    Function FindLevelModel(pkg As Package) As ExportTableItem
        ' This isn't perfect, but should work most of the time.
        Dim candidates As New List(Of ExportTableItem)

        For Each item In pkg.ExportTable
            If StrEq(item.ClassNameStr, "Model") Then
                If (item.ObjectFlags = (EObjectFlags.RF_Transactional Or EObjectFlags.RF_LoadForClient Or EObjectFlags.RF_LoadForServer Or EObjectFlags.RF_LoadForEdit)) Then
                    If StrEq(item.ObjectNameStr, "Model1") Then
                        Return item
                    End If
                    candidates.Add(item)
                End If
            End If
        Next

        ' Try another heuristic...
        Dim maxSize = 0
        Dim maxItem As ExportTableItem = Nothing
        For i = 0 To candidates.Count - 1
            If candidates(i).OriginalSize > maxSize Then
                maxSize = candidates(i).OriginalSize
                maxItem = candidates(i)
            End If
        Next

        Return maxItem
    End Function

    Function IsALevel(pkg As Package) As Boolean
        For Each item In pkg.ExportTable
            If StrEq(item.ObjectNameStr, "MyLevel") And
               StrEq(item.ClassNameStr, "Level") Then
                Return True
            End If
        Next
        Return False
    End Function

    Sub SaveLevel(pkg As Package, t3dfile As String, bspfile As String)
        Dim w As New IO.StreamWriter(t3dfile)

        Dim actors = pkg.GetActorList()

        Dim modelExp = FindLevelModel(pkg)
        Dim model As Model = Nothing
        If modelExp IsNot Nothing Then
            model = GetModel(modelExp)
        Else
            Console.ForegroundColor = ConsoleColor.Red
            Console.WriteLine("Level model not found.")
            Console.ResetColor()
        End If

        w.WriteLine("Begin Map")

        For i = 0 To actors.GetUpperBound(0)
            If actors(i) <> 0 Then
                w.Write(ActorToT3D(pkg.ExportTable(actors(i) - 1)))
            End If
        Next

        w.WriteLine("End Map")

        w.Close()

        SaveLevelModel(model, bspfile)
    End Sub

    Function FlipVector(vect As Vector) As Vector
        Return New Vector(-vect.X, -vect.Y, -vect.Z)
    End Function

    Sub WriteLevelModel(mdl As Model, stream As IO.StreamWriter)
        stream.WriteLine("Begin Actor Class=Brush Name=Brush0")
        stream.WriteLine("MainScale=(SheerAxis=SHEER_ZX)")
        stream.WriteLine("PostScale=(SheerAxis=SHEER_ZX)")
        stream.WriteLine("Group=""Cube""")

        Dim str As New System.Text.StringBuilder

        stream.WriteLine("Begin Brush Name=" & mdl.Name)
        stream.WriteLine("Begin PolyList")

        For Each n In mdl.Nodes
            If n.NumVertices > 0 Then
                Dim surf = mdl.Surfs(n.iSurf)

                stream.Write("Begin Polygon")
                If surf.Texture <> 0 Then stream.Write(" Texture=" & mdl.Level.GetFullObjectName(surf.Texture, False))
                If surf.PolyFlags <> 0 Then stream.Write(" Flags=" & surf.PolyFlags)
                stream.WriteLine()

                stream.WriteLine("Origin " & mdl.Points(surf.pBase).ToT3DString())
                stream.WriteLine("Normal " & FlipVector(mdl.Vectors(surf.vNormal)).ToT3DString())
                stream.WriteLine("TextureU " & mdl.Vectors(surf.vTextureU).ToT3DString())
                stream.WriteLine("TextureV " & mdl.Vectors(surf.vTextureV).ToT3DString())
                If surf.PanU <> 0 Or surf.PanV <> 0 Then
                    stream.WriteLine("Pan U=" & surf.PanU & " V=" & surf.PanV)
                End If

                'For i = n.iVertPool To n.iVertPool + n.NumVertices - 1
                For i = n.iVertPool + n.NumVertices - 1 To n.iVertPool Step -1 ' Flip the normals.
                    stream.WriteLine("Vertex " & mdl.Points(mdl.Verts(i).pVertex).ToT3DString)
                Next

                stream.WriteLine("End Polygon")
            End If
        Next

        stream.WriteLine("End PolyList")
        stream.WriteLine("End Brush")

        stream.WriteLine("Brush=Model'MyLevel." & mdl.Name & "'")
        stream.WriteLine("End Actor")
    End Sub

    Function ActorToT3D(act As ExportTableItem) As String
        Dim pkg = act.Pkg
        Dim text As New System.Text.StringBuilder

        Dim props = act.GetProperties
        Dim modelNum = props.GetInt("Brush")

        ' Don't export static brushes.
        If modelNum <> 0 And StrEq(act.ClassNameStr, "Brush") Then
            Return ""
        End If

        text.Append("Begin Actor Class=")
        text.Append(act.ClassNameStr)
        text.Append(" Name=")
        text.AppendLine(act.ObjectNameStr)

        If modelNum <> 0 Then
            Dim model = pkg.ExportTable(modelNum - 1)
            text.Append("Begin Brush Name=")
            text.AppendLine(model.ObjectNameStr)
            'Dim polysNum = Models.GetPolysNumFromModel(model)
            Dim polysNum = GetModel(model).Polys
            If polysNum <> 0 Then
                text.AppendLine(Models.GetPolys(pkg.ExportTable(polysNum - 1)).ToString)
            Else
                text.AppendLine("; ERROR: Can't find Polys")
            End If
            text.AppendLine("End Brush")
        End If

        For Each prop In props.ToStringList
            Select Case prop.Name.ToLower
                Case "bselected", "level", "oldlocation", "region", "xlevel"
                    ' Don't write those
                Case Else
                    text.AppendLine(prop.Name & "=" & prop.Value)
            End Select
        Next

        text.AppendLine("End Actor")

        Return text.ToString

    End Function

    Sub SaveLevelModel(mdl As Model, file As String)
        Using w As New UnBinaryWriter(New IO.FileStream(file, IO.FileMode.Create, IO.FileAccess.Write))
            w.WriteInt32(24) 'Type (Level model)

            Dim pkg = mdl.Level
            w.WriteInt32(pkg.ExportTable.Count)
            For Each item In pkg.ExportTable
                w.WriteZString(item.ClassNameStr)
                w.WriteZString(item.ObjectNameStr)
            Next

            w.WriteInt32(mdl.Vectors.Count)
            For Each v In mdl.Vectors
                w.WriteVector(v)
            Next

            w.WriteInt32(mdl.Points.Count)
            For Each v In mdl.Points
                w.WriteVector(v)
            Next

            w.WriteInt32(mdl.Verts.Count)
            For Each v In mdl.Verts
                w.WriteInt32(v.pVertex)
            Next

            w.WriteInt32(mdl.Lights.Count)
            For Each light In mdl.Lights
                If light < 0 Or light > pkg.ExportTable.Count Then
                    WriteRed("ERROR: Invalid light reference")
                    Return
                End If
                w.WriteInt32(light)
            Next

            w.WriteInt32(mdl.LightBits.Length)
            w.Write(mdl.LightBits)

            w.WriteInt32(mdl.LightMaps.Count)
            For Each lm In mdl.LightMaps
                If lm.DataOffset < 0 Or lm.DataOffset > mdl.LightBits.Length Then
                    WriteRed("ERROR: Invalid LightMap.DataOffset")
                    Return
                ElseIf lm.DataOffset = mdl.LightBits.Length Then '???
                    Console.ForegroundColor = ConsoleColor.Yellow
                    Console.WriteLine("Suspicious LightMap.DataOffset")
                    Console.ResetColor()
                    w.WriteInt32(-1)
                Else
                    w.WriteInt32(lm.DataOffset)
                End If
                w.WriteVector(lm.Pan)
                w.WriteInt32(lm.UClamp)
                w.WriteInt32(lm.VClamp)
                w.WriteSingle(lm.UScale)
                w.WriteSingle(lm.VScale)
                If lm.iLightActors < -1 Or lm.iLightActors >= mdl.Lights.Count Then
                    WriteRed("ERROR: Invalid LightMap.iLightActors")
                    Return
                End If
                w.WriteInt32(lm.iLightActors)
            Next

            w.WriteInt32(mdl.Surfs.Count)
            For Each surf In mdl.Surfs
                If surf.Texture = 0 Then
                    w.WriteZString("")
                Else
                    w.WriteZString(mdl.Level.GetFullObjectName(surf.Texture, False))
                End If
                w.WriteInt32(surf.PolyFlags)
                If surf.iLightMap < -1 Or surf.iLightMap >= mdl.LightMaps.Count Then
                    WriteRed("ERROR: Invalid BspSurf.iLightMap")
                    Return
                End If
                w.WriteInt32(surf.iLightMap)
                If surf.pBase < 0 Or surf.pBase >= mdl.Points.Count Then
                    WriteRed("ERROR: Invalid BspSurf.pBase")
                    Return
                End If
                w.WriteVector(mdl.Points(surf.pBase))
                If surf.vNormal < 0 Or surf.vNormal >= mdl.Vectors.Count Then
                    WriteRed("ERROR: Invalid BspSurf.vNormal")
                    Return
                End If
                w.WriteVector(FlipVector(mdl.Vectors(surf.vNormal)))
                If surf.vTextureU < 0 Or surf.vTextureU >= mdl.Vectors.Count Then
                    WriteRed("ERROR: Invalid BspSurf.vTextureU")
                    Return
                End If
                w.WriteVector(mdl.Vectors(surf.vTextureU))
                If surf.vTextureV < 0 Or surf.vTextureV >= mdl.Vectors.Count Then
                    WriteRed("ERROR: Invalid BspSurf.vTextureV")
                    Return
                End If
                w.WriteVector(mdl.Vectors(surf.vTextureV))
                w.WriteInt16(surf.PanU)
                w.WriteInt16(surf.PanV)
            Next

            w.WriteInt32(mdl.Nodes.Count)
            For Each node In mdl.Nodes
                w.WriteSingle(node.Plane.X)
                w.WriteSingle(node.Plane.Y)
                w.WriteSingle(node.Plane.Z)
                w.WriteSingle(node.Plane.W)
                w.Write(node.ZoneMask)
                w.WriteByte(node.NodeFlags)
                w.WriteInt32(node.iVertPool)

                If node.iSurf < 0 Or node.iSurf >= mdl.Surfs.Count Then
                    WriteRed("ERROR: Invalid BspNode.iSurf")
                    Return
                End If
                w.WriteInt32(node.iSurf)

                w.WriteInt32(node.iChild0)
                w.WriteInt32(node.iChild1)
                w.WriteInt32(node.iChild2)
                w.WriteByte(node.iZone0)
                w.WriteByte(node.iZone1)
                w.WriteByte(node.NumVertices)
                w.WriteInt32(node.iLeaf0)
                w.WriteInt32(node.iLeaf1)
            Next

            w.WriteInt32(mdl.Zones.Count)
            For Each zone In mdl.Zones
                If zone.ZoneActor < 0 Or zone.ZoneActor > pkg.ExportTable.Count Then
                    WriteRed("ERROR: Invalid ZoneInfo reference")
                    Return
                End If
                w.Write(zone.ZoneActor)
            Next

            w.WriteInt32(mdl.Leaves.Count)
            For Each lvz In mdl.Leaves
                w.Write(lvz.VisibleZones)
            Next

            'New (2021):

            w.WriteInt32(mdl.LeafHulls.Size / 4)
            w.Write(mdl.LeafHulls.ToArray)

            For Each node In mdl.Nodes
                w.Write(node.iCollisionBound)
            Next

        End Using
    End Sub
End Module
